/*******************************************************************************
* Copyright (C) 2024 Intel Corporation
*
* This software and the related documents are Intel copyrighted  materials,  and
* your use of  them is  governed by the  express license  under which  they were
* provided to you (License).  Unless the License provides otherwise, you may not
* use, modify, copy, publish, distribute,  disclose or transmit this software or
* the related documents without Intel's prior written permission.
*
* This software and the related documents  are provided as  is,  with no express
* or implied  warranties,  other  than those  that are  expressly stated  in the
* License.
*******************************************************************************/

/*
*
*  Content:
*       This example demonstrates use of oneAPI Math Kernel Library (oneMKL)
*       DPCPP USM API oneapi::mkl::sparse::matmatd to perform general
*       sparse matrix-sparse matrix multiplication with a dense result matrix
*       on a SYCL device (GPU). This example uses sparse matrices in CSR format.
*
*           C = alpha * op(A) * op(B) + beta * C
*
*       where op() is defined by one of
*           oneapi::mkl::transpose::{nontrans,trans,conjtrans}
*
*       The supported floating point data types for matmatd matrix data are:
*           float
*           double
*           std::complex<float>
*           std::complex<double>
*
*       The supported matrix formats for matmatd are:
*           CSR
*
*******************************************************************************/

// stl includes
#include <algorithm>
#include <cstdlib>
#include <cstring>
#include <iostream>
#include <iterator>
#include <limits>
#include <list>
#include <vector>

#include "mkl.h"
#include "oneapi/mkl.hpp"
#include <sycl/sycl.hpp>

// local includes
#include "common_for_examples.hpp"
#include "./include/common_for_sparse_examples.hpp"

//
// Main example for Sparse Matrix-Sparse Matrix Multiply with a dense result,
// consisting of initialization of A and B matrices through process of
// creating the dense matrix C as the product
//
// C = alpha op(A) * op(B) + beta C
//
// In this case, the square symmetric matrices A and B are generated separately
// occupying different memory, but with the same structure and values, so that
// in the end C is the square of A.
//
template <typename dataType, typename intType>
int run_sparse_blas_example(sycl::queue &q)
{
    bool good = true;

    oneapi::mkl::sparse::matrix_handle_t csrA = nullptr;
    oneapi::mkl::sparse::matrix_handle_t csrB = nullptr;

    //
    // array memory management tools
    //
    std::vector<intType *> int_ptr_vec;
    std::vector<dataType *> data_ptr_vec;
    std::vector<std::int64_t *> i64_ptr_vec;
    std::vector<void *> void_ptr_vec;

    try {

        // Initialize data for Sparse Matrix - Sparse Matrix Multiply
        oneapi::mkl::transpose opA = oneapi::mkl::transpose::nontrans;
        oneapi::mkl::transpose opB = oneapi::mkl::transpose::nontrans;

        oneapi::mkl::index_base a_index = oneapi::mkl::index_base::zero;
        oneapi::mkl::index_base b_index = oneapi::mkl::index_base::one;

        //
        // set up dimensions of matrix products
        //
        intType size = 4;

        const std::int64_t a_nrows = size * size * size;
        const std::int64_t a_ncols = a_nrows;
              std::int64_t a_nnz   = 27 * a_nrows;
        const std::int64_t b_nrows = size * size * size;
        const std::int64_t b_ncols = b_nrows;
              std::int64_t b_nnz   = 27 * b_nrows;
        const std::int64_t c_nrows = size * size * size;
        const std::int64_t c_ncols = c_nrows;

        //
        // setup A data locally in CSR format
        //
        intType *a_rowptr_host = sycl::malloc_host<intType>(a_nrows + 1, q);
        intType *a_colind_host = sycl::malloc_host<intType>(a_nnz, q);
        dataType *a_values_host  = sycl::malloc_host<dataType>(a_nnz, q);
        if (!a_rowptr_host || !a_colind_host || !a_values_host) {
            std::string errorMessage =
                "Failed to allocate USM host memory arrays \n"
                " for CSR A matrix: a_rowptr(" + std::to_string((a_nrows+1)*sizeof(intType)) + " bytes)\n"
                "                   a_colind(" + std::to_string((a_nnz)*sizeof(intType)) + " bytes)\n"
                "                   a_values(" + std::to_string((a_nnz)*sizeof(dataType)) + " bytes)";

            throw std::runtime_error(errorMessage);
        }
        int_ptr_vec.push_back(a_rowptr_host);
        int_ptr_vec.push_back(a_colind_host);
        data_ptr_vec.push_back(a_values_host);

        intType a_ind = a_index == oneapi::mkl::index_base::zero ? 0 : 1;
        generate_sparse_matrix<dataType, intType>(size, a_rowptr_host, a_colind_host, a_values_host, a_ind);
        a_nnz = a_rowptr_host[a_nrows] - a_ind;

        intType *a_rowptr = sycl::malloc_device<intType>(a_nrows + 1, q);
        intType *a_colind = sycl::malloc_device<intType>(a_nnz, q);
        dataType *a_values  = sycl::malloc_device<dataType>(a_nnz, q);

        if (!a_rowptr || !a_colind || !a_values) {
           std::string errorMessage =
               "Failed to allocate USM device memory arrays \n"
               " for CSR A matrix: a_rowptr(" + std::to_string((a_nrows+1)*sizeof(intType)) + " bytes)\n"
               "                   a_colind(" + std::to_string((a_nnz)*sizeof(intType)) + " bytes)\n"
               "                   a_values(" + std::to_string((a_nnz)*sizeof(dataType)) + " bytes)";

            throw std::runtime_error(errorMessage);
        }
        int_ptr_vec.push_back(a_rowptr);
        int_ptr_vec.push_back(a_colind);
        data_ptr_vec.push_back(a_values);

        // copy A matrix USM data from host to device
        auto ev_cpy_ia = q.copy<intType>(a_rowptr_host, a_rowptr, a_nrows + 1);
        auto ev_cpy_ja = q.copy<intType>(a_colind_host, a_colind, a_nnz);
        auto ev_cpy_a  = q.copy<dataType>(a_values_host, a_values, a_nnz);

        //
        // setup B data locally in CSR format
        //
        intType *b_rowptr_host = sycl::malloc_host<intType>(b_nrows + 1, q);
        intType *b_colind_host = sycl::malloc_host<intType>(b_nnz, q);
        dataType *b_values_host  = sycl::malloc_host<dataType>(b_nnz, q);
        if (!b_rowptr_host || !b_colind_host || !b_values_host) {
            std::string errorMessage =
                "Failed to allocate USM host memory arrays \n"
                " for CSR B matrix: b_rowptr(" + std::to_string((b_nrows+1)*sizeof(intType)) + " bytes)\n"
                "                   b_colind(" + std::to_string((b_nnz)*sizeof(intType)) + " bytes)\n"
                "                   b_values(" + std::to_string((b_nnz)*sizeof(dataType)) + " bytes)";

            throw std::runtime_error(errorMessage);
        }
        int_ptr_vec.push_back(b_rowptr_host);
        int_ptr_vec.push_back(b_colind_host);
        data_ptr_vec.push_back(b_values_host);

        intType b_ind = b_index == oneapi::mkl::index_base::zero ? 0 : 1;
        generate_sparse_matrix<dataType, intType>(size, b_rowptr_host, b_colind_host, b_values_host, b_ind);
        b_nnz = b_rowptr_host[b_nrows] - b_ind;

        intType *b_rowptr = sycl::malloc_device<intType>(b_nrows + 1, q);
        intType *b_colind = sycl::malloc_device<intType>(b_nnz, q);
        dataType *b_values  = sycl::malloc_device<dataType>(b_nnz, q);
        if (!b_rowptr || !b_colind || !b_values) {
            std::string errorMessage =
                "Failed to allocate USM device memory arrays \n"
                " for CSR B matrix: b_rowptr(" + std::to_string((b_nrows+1)*sizeof(intType)) + " bytes)\n"
                "                   b_colind(" + std::to_string((b_nnz)*sizeof(intType)) + " bytes)\n"
                "                   b_values(" + std::to_string((b_nnz)*sizeof(dataType)) + " bytes)";

            throw std::runtime_error(errorMessage);
        }
        int_ptr_vec.push_back(b_rowptr);
        int_ptr_vec.push_back(b_colind);
        data_ptr_vec.push_back(b_values);

        //
        // setup C data locally in dense matrix format
        //
        dataType alpha = dataType(1);
        dataType beta = dataType(0);
        auto c_layout = oneapi::mkl::layout::row_major;
        intType ldc = (c_layout == oneapi::mkl::layout::row_major) ? c_ncols : c_nrows;
        intType c_size = (c_layout == oneapi::mkl::layout::row_major) ? ldc * c_nrows : ldc * c_ncols;
        auto dnsC = sycl::malloc_device<dataType>(c_size, q);

        if (!dnsC) {
            std::string errorMessage =
                "Failed to allocate USM device memory arrays \n"
                " for dense C matrix: dnsC(" + std::to_string((c_size)*sizeof(dataType)) + " bytes)";

            throw std::runtime_error(errorMessage);
        }
        data_ptr_vec.push_back(dnsC);

        // copy B matrix USM data from host to device
        auto ev_cpy_ib = q.copy<intType>(b_rowptr_host, b_rowptr, b_nrows + 1);
        auto ev_cpy_jb = q.copy<intType>(b_colind_host, b_colind, b_nnz);
        auto ev_cpy_b  = q.copy<dataType>(b_values_host, b_values, b_nnz);

        //
        // Execute Matrix Multiply
        //

        std::cout << "\n\t\tsparse::matmatd parameters:\n";
        std::cout << "\t\t\topA = " << opA << std::endl;
        std::cout << "\t\t\topB = " << opB << std::endl;

        std::cout << "\t\t\tA_nrows = A_ncols = " << a_nrows << std::endl;
        std::cout << "\t\t\tB_nrows = B_ncols = " << b_nrows << std::endl;
        std::cout << "\t\t\tC_nrows = C_ncols = " << c_nrows << std::endl;

        std::cout << "\t\t\tA_index = " << a_index << std::endl;
        std::cout << "\t\t\tB_index = " << b_index << std::endl;

        oneapi::mkl::sparse::init_matrix_handle(&csrA);
        oneapi::mkl::sparse::init_matrix_handle(&csrB);

        auto ev_setA = oneapi::mkl::sparse::set_csr_data(q, csrA, a_nrows, a_ncols, a_nnz, a_index,
                a_rowptr, a_colind, a_values, {ev_cpy_ia, ev_cpy_ja, ev_cpy_a});
        auto ev_setB = oneapi::mkl::sparse::set_csr_data(q, csrB, b_nrows, b_ncols, b_nnz, b_index,
                b_rowptr, b_colind, b_values, {ev_cpy_ib, ev_cpy_jb, ev_cpy_b});
        auto ev_fillC = q.fill<dataType>(dnsC, dataType(0), c_size);

        //
        // compute sparse matrix-matrix product with dense result
        //
        auto ev2 = oneapi::mkl::sparse::matmatd(
            q, c_layout, opA, opB, alpha, csrA, csrB, beta, dnsC, c_nrows, c_ncols,
            ldc, {ev_setA, ev_setB, ev_fillC});

        // Copy C matrix to host for printing
        const intType c_nrows_copy = std::min<std::int64_t>(2, c_nrows); // only copy over this many rows of C to host

        dataType *dnsC_host = sycl::malloc_host<dataType>(c_ncols * c_nrows_copy, q);
        if (!dnsC_host) {
            std::string errorMessage =
                "Failed to allocate USM host memory arrays \n"
                " for dense C matrix: dnsC(" + std::to_string((c_ncols * c_nrows_copy)*sizeof(dataType)) + " bytes)";

            throw std::runtime_error(errorMessage);
        }
        data_ptr_vec.push_back(dnsC_host);

        auto ev_copy_c = q.submit([&](sycl::handler& cgh) {
            cgh.depends_on(ev2);
            if (c_layout == oneapi::mkl::layout::row_major) {
                cgh.parallel_for(
                    sycl::range<2>(c_nrows_copy, c_ncols),
                    [=](sycl::item<2> item) {
                        dnsC_host[c_ncols * item.get_id(0) + item.get_id(1)] =
                            dnsC[ldc * item.get_id(0) + item.get_id(1)];
                    });
            }
            else if (c_layout == oneapi::mkl::layout::col_major) {
                cgh.parallel_for(
                    sycl::range<2>(c_ncols, c_nrows_copy),
                    [=](sycl::item<2> item) {
                        dnsC_host[c_nrows * item.get_id(0) + item.get_id(1)] =
                            dnsC[ldc * item.get_id(0) + item.get_id(1)];
                    });
            }
            else {
                throw std::runtime_error("Bad layout provided for C matrix");
            }
        });
        ev_copy_c.wait(); // make sure copy is done before reading from it

        // print out a portion of C solution
        sycl::event ev_print = q.submit([&](sycl::handler &cgh) {
            cgh.depends_on({ev_copy_c});
            auto kernel = [=]() {
                std::cout << "C matrix [first " << c_nrows_copy << " rows]:" << std::endl;
                for (intType row = 0; row < c_nrows_copy; ++row) {
                    for (intType col = 0; col < c_ncols; ++col) {
                        dataType val = (c_layout == oneapi::mkl::layout::row_major) ?
                            dnsC_host[ldc*row + col] : dnsC_host[ldc*col + row];
                        std::cout << "C(" << row << ", " << col << ") = " << val
                                  << std::endl;
                    }
                }
            };
            cgh.host_task(kernel);
        });

        auto ev_relA = oneapi::mkl::sparse::release_matrix_handle(q, &csrA, {ev_print});
        auto ev_relB = oneapi::mkl::sparse::release_matrix_handle(q, &csrB, {ev_print});

        ev_relA.wait();
        ev_relB.wait();

        q.wait_and_throw();

        good = true;
    }
    catch (oneapi::mkl::unimplemented const &e) {
        std::cout << "\t\tCaught oneMKL unimplemented exception:\n" << e.what() << std::endl;
        good = true;
    }
    catch (sycl::exception const &e) {
        std::cout << "\t\tCaught synchronous SYCL exception:\n" << e.what() << std::endl;
        good = false;
    }
    catch (std::exception const &e) {
        std::cout << "\t\tCaught std exception:\n" << e.what() << std::endl;
        good = false;
    }

    q.wait();

    // backup cleaning of matrix handle and others for if exceptions happened
    oneapi::mkl::sparse::release_matrix_handle(q, &csrA, {}).wait();
    oneapi::mkl::sparse::release_matrix_handle(q, &csrB, {}).wait();

    cleanup_arrays<dataType, intType>(data_ptr_vec, int_ptr_vec, i64_ptr_vec, void_ptr_vec, q);

    return good ? 0 : 1;
}

//
// Description of example setup, apis used and supported floating point type
// precisions
//
void print_example_banner()
{

    std::cout << "" << std::endl;
    std::cout << "###############################################################"
                 "#########"
              << std::endl;
    std::cout << "# Sparse Matrix-Sparse Matrix Multiply with dense result Example: " << std::endl;
    std::cout << "# " << std::endl;
    std::cout << "#    C = alpha * op(A) * op(B) + beta * C" << std::endl;
    std::cout << "# " << std::endl;
    std::cout << "# where A and B are sparse matrices in CSR format, and C is the\n"
                 "# dense matrix product"
              << std::endl;
    std::cout << "# " << std::endl;
    std::cout << "# Using apis:" << std::endl;
    std::cout << "#   sparse::matmatd" << std::endl;
    std::cout << "# " << std::endl;
    std::cout << "#   sparse::init_matrix_handle" << std::endl;
    std::cout << "#   sparse::set_csr_data" << std::endl;
    std::cout << "#   sparse::release_matrix_handle" << std::endl;
    std::cout << "# " << std::endl;
    std::cout << "# Supported floating point type precisions:" << std::endl;
    std::cout << "#   float" << std::endl;
    std::cout << "#   double" << std::endl;
    std::cout << "#   std::complex<float>" << std::endl;
    std::cout << "#   std::complex<double>" << std::endl;
    std::cout << "# " << std::endl;
    std::cout << "###############################################################"
                 "#########"
              << std::endl;
    std::cout << std::endl;
}

//
// Main entry point for example.
//
// Dispatches to appropriate device types as set at build time with flag:
// -DSYCL_DEVICES_cpu -- only runs SYCL CPU implementation
// -DSYCL_DEVICES_gpu -- only runs SYCL GPU implementation
// -DSYCL_DEVICES_all (default) -- runs on all: cpu and gpu devices
//
//  For each device selected and each supported data type,
//  run_sparse_blas_example() is run with all supported data types,
//  if any fail, we move on to the next device.
//

int main(int argc, char **argv)
{
    print_example_banner();

    std::list<my_sycl_device_types> list_of_devices;
    set_list_of_devices(list_of_devices);

    int status = 0;
    for (auto it = list_of_devices.begin(); it != list_of_devices.end(); ++it) {
        try {
            sycl::device my_dev;
            bool my_dev_is_found = false;
            get_sycl_device(my_dev, my_dev_is_found, *it);

            if (my_dev_is_found) {
                std::cout << "Running tests on " << sycl_device_names[*it] << ".\n";

                // Catch asynchronous exceptions
                auto exception_handler = [](sycl::exception_list exceptions) {
                    for (std::exception_ptr const &e : exceptions) {
                        try {
                            std::rethrow_exception(e);
                        }
                        catch (sycl::exception const &e) {
                            std::cout << "Caught asynchronous SYCL exception: \n"
                                << e.what() << std::endl;
                        }
                    }
                };

                sycl::queue q(my_dev, exception_handler);

                std::cout << "\tRunning with single precision real data type:" << std::endl;
                status |= run_sparse_blas_example<float, std::int32_t>(q);

                if (my_dev.get_info<sycl::info::device::double_fp_config>().size() != 0) {
                    std::cout << "\tRunning with double precision real data type:" << std::endl;
                    status |= run_sparse_blas_example<double, std::int32_t>(q);
                }

                std::cout << "\tRunning with single precision complex data type:" << std::endl;
                status |= run_sparse_blas_example<std::complex<float>, std::int32_t>(q);

                if (my_dev.get_info<sycl::info::device::double_fp_config>().size() != 0) {
                    std::cout << "\tRunning with double precision complex data type:" << std::endl;
                    status |= run_sparse_blas_example<std::complex<double>, std::int32_t>(q);
                }

            }
            else {
#ifdef FAIL_ON_MISSING_DEVICES
                std::cout << "No " << sycl_device_names[*it]
                    << " devices found; Fail on missing devices "
                    "is enabled.\n";
                return 1;
#else
                std::cout << "No " << sycl_device_names[*it] << " devices found; skipping "
                    << sycl_device_names[*it] << " tests.\n";
#endif
            }
        }
        catch (sycl::exception const &e) {
            std::cout << "\t\tCaught SYCL exception at driver level: \n" << e.what() << std::endl;
            continue; // stop with device, but move on to other devices
        }
        catch (std::exception const &e) {
            std::cout << "\t\tCaught std exception at driver level: \n" << e.what() << std::endl;
            continue; // stop with device, but move on to other devices
        }

    } // for device

    mkl_free_buffers();
    return status;
}
